#include <stdio.h>

#include <osmocom/core/select.h>
#include <osmocom/core/utils.h>

#include <osmocom/core/logging.h>
#include <osmocom/core/application.h>
#include <osmocom/core/fsm.h>
#include <osmocom/vty/vty.h>
#include <osmocom/vty/telnet_interface.h>

#include <osmocom/sigtran/osmo_ss7.h>
#include <osmocom/sigtran/sccp_sap.h>
#include <osmocom/sigtran/sccp_helpers.h>
#include <osmocom/sigtran/protocol/sua.h>
#include <osmocom/sigtran/protocol/m3ua.h>

#include <osmocom/gsm/protocol/gsm_08_08.h>
#include <osmocom/gsm/gsm0808.h>
#include <osmocom/gsm/gsm0808_utils.h>

/*
 * FIXME: This is stuff we need to handle the SCCP/M3UA/SCTP connection,
 * this needs to be packed somwhere. By having it global now, we also
 * only support one connection (which is ok for tests)
 */

#define SSN_BSSAP	254 /* SCCP_SSN_BSSAP */
#define SENDER_PC	1
#define RECEIVER_PC	23

struct osmo_sccp_user *scu = NULL;
struct osmo_sccp_instance *sccp = NULL;



/*
 * Logging sutff
 */

/* Logging */
static const struct log_info_cat log_info_cat[] = {
};

/* Logging */
static const struct log_info log_info = {
	.cat = log_info_cat,
	.num_cat = ARRAY_SIZE(log_info_cat),
};

/* Initalize logging */
static void init_logging(void)
{
	const int log_cats[] = { DLSS7, DLSUA, DLM3UA, DLSCCP, DLINP };
	unsigned int i;

	osmo_init_logging(&log_info);

	for (i = 0; i < ARRAY_SIZE(log_cats); i++)
		log_set_category_filter(osmo_stderr_target, log_cats[i], 1, LOGL_DEBUG);
}





/*
 * BSSAP Drafts
 */


/* Handle BSSMAP reset */
static void bssmap_handle_reset(struct osmo_scu_prim *scu_prim, const uint8_t *data, unsigned int data_len)
{
	/* FIXME: Tell lower layers that the BSC was reset! */
	struct msgb *msg = NULL;

	printf("Received RESET from BSC\n");

	msg = gsm0808_create_reset_ack();
	osmo_sccp_tx_unitdata(scu, &scu_prim->u.unitdata.called_addr, &scu_prim->u.unitdata.calling_addr, msg->data, msg->len);

	msgb_free(msg);
}

/* Handle UNITDATA BSSMAP messages */
static void bssmap_rcvmsg_udt(struct osmo_scu_prim *scu_prim, const uint8_t *data, unsigned int data_len)
{

	/* Note: When in the MSC role, RESET ACK is the only valid message that
	 * can be received via UNITDATA */

	if (data_len < 1) {
		printf("Error: No data received -- discarding message!\n");
		return;
	}

	printf("Rx MSC UDT BSSMAP %s\n", gsm0808_bssmap_name(data[0]));

	switch (data[0]) {
	case BSS_MAP_MSG_RESET:
		bssmap_handle_reset(scu_prim, data, data_len);
		break;
	default:
		printf("Unimplemented message format: %s -- message discarded!\n", gsm0808_bssmap_name(data[0]));
	}
}

/* Handle incoming UNITDATA */
static void bsc_handle_udt(struct osmo_scu_prim *scu_prim, const uint8_t *data, unsigned int data_len)
{
	/* Note: The only valid message type that can be received
	 * via UNITDATA are BSS Management messages */
	struct bssmap_header *bs;

	printf("Rx BSC UDT: %s\n", osmo_hexdump(data,  data_len));

	if (data_len < sizeof(*bs)) {
		printf("Error: Header is too short!\n");
		return;
	}

	bs = (struct bssmap_header *) data;
	if (bs->length < data_len - sizeof(*bs)) {
		printf("Error: Message is too short!\n");
		return;
	}

	switch (bs->type) {
	case BSSAP_MSG_BSS_MANAGEMENT:
		data += sizeof(*bs);
		bssmap_rcvmsg_udt(scu_prim, data, data_len);
		break;
	default:
		printf("Error: Unimplemented message type: %s -- message discarded!\n", gsm0808_bssmap_name(bs->type));
	}
}



static void bsc_clear_conn(struct osmo_scu_prim *scu_prim)
{
	struct msgb *msg = NULL;

	printf("Sending CLEAR COMMAND to BSC...\n");

	msg = gsm0808_create_clear_command(1);

	osmo_sccp_tx_data(scu, scu_prim->u.data.conn_id, msg->data, msg->len);

	msgb_free(msg);
}



/* a simple SCCP User which accepts all connections and echos back all
 * DATA + UNITDATA */
static int echo_prim_cb(struct osmo_prim_hdr *oph, void *_scu)
{
	struct osmo_sccp_user *scu = _scu;
	struct osmo_scu_prim *scu_prim = (struct osmo_scu_prim *) oph;
	const uint8_t *data = msgb_l2(oph->msg);
	unsigned int data_len = msgb_l2len(oph->msg);


	printf("###### echo_prim_cb()!\n");

	switch (OSMO_PRIM_HDR(&scu_prim->oph)) {
	case OSMO_PRIM(OSMO_SCU_PRIM_N_CONNECT, PRIM_OP_INDICATION):
		printf("%s: Accepting N-CONNECT.ind (local_ref=%u)\n",
			__func__, scu_prim->u.connect.conn_id);
		osmo_sccp_tx_conn_resp(scu, scu_prim->u.connect.conn_id,
				     &scu_prim->u.connect.called_addr,
				     NULL, 0);
		break;
	case OSMO_PRIM(OSMO_SCU_PRIM_N_DATA, PRIM_OP_INDICATION):
	//	printf("%s: Echoing N-DATA.ind (local_ref=%u)\n",
	//		__func__, scu_prim->u.data.conn_id);
	//	osmo_sccp_tx_data(scu, scu_prim->u.data.conn_id,
	//			  data, data_len);
		bsc_clear_conn(scu_prim);
		break;
	case OSMO_PRIM(OSMO_SCU_PRIM_N_UNITDATA, PRIM_OP_INDICATION):
		/* Forward UNITDATA to BSSMAP */
		bsc_handle_udt(scu_prim, data, data_len);
		break;
	default:
		printf("%s: Unknown primitive %u:%u\n", __func__,
			oph->primitive, oph->operation);
		break;
	}

	return 0;
}

int main(void)
{





	init_logging();
	osmo_ss7_init();


	sccp = osmo_sccp_simple_server(NULL, SENDER_PC, OSMO_SS7_ASP_PROT_M3UA,
					-1, "127.0.0.2");

	/* Why? */
	osmo_sccp_simple_server_add_clnt(sccp, OSMO_SS7_ASP_PROT_M3UA,
					"RECEIVER", RECEIVER_PC, -1, 0, NULL);

	scu = osmo_sccp_user_bind(sccp, "echo", &echo_prim_cb, SSN_BSSAP);
	printf("###### osmo_sccp_user_bind() done!\n");


	printf("###### Enter main loop!\n");
	while (1) {
		osmo_select_main(0);
	}

	return 0;
}
